--[[
Rewritten from Incursion Guild Alts, originally by Zorlan.  See there for the
full tortured history.

]]

local nametag, addon = ...
local L = LibStub("AceLocale-3.0"):GetLocale(nametag)

local altrank = L["alt"]:lower()   -- must be lowercase

addon.defaults = {
	profile = {
		enable = true,
		use_alt_rank = false,
		stats_at_login = true,
		alt_pattern = "[Aa][lL][tT][ ]?%(<NAME>%)",
		tag_channels = "raid",
		do_rank_tags = false,
		do_rank_tags_officer = false,
		debug = false,
		hcd_debug = false,
		debugging_frame = 1,
		hcd_debugging_frame = 2,
		max_chatframe_num = 7,
	},
	global = {
		guilds = { },
	},
}


local pattern_editing_safe
addon.options = {
	name = "Ouro Guild Alts",
	type = 'group',
	childGroups = 'tab',
	handler = addon,    -- functions listed as strings called as addon:func
	get = "GetOption",
	set = "SetOption",
	args = {
		general = {
			name = GENERAL,
			desc = L["General options"],
			type = 'group',
			order = 10,
			args = {
				version = {
					--name = filled in during OnInit
					type = 'description',
					fontSize = "large",
					cmdHidden = true,
					width = 'full',
					order = 1,
				},
				enable = {
					name = ENABLE,
					desc = L["Use this addon"],
					type = 'toggle',
					arg  = "ToggleEnable",
					order = 5,
				},
				rescan = {
					name = L["Rescan Guild Roster"],
					desc = L["Reloads the list of alts from the guild roster"],
					type = 'execute',
					func = "LoadGuild",
					order = 10,
				},
				--[[spacer = {
					name = '',
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = 14,
				},]]
				use_alt_rank = {
					name = L["Check Ranks"],
					--desc = -- filled in below,
					type = 'toggle',
					width = 'double',
					order = 15,
				},
				stats_at_login = {
					name = L["Stats at Login"],
					desc = L["During login, print statistics about the count of alts and mains detected in the guild."],
					type = 'toggle',
					width = 'double',
					order = 20,
				},
				instructions = {
					name = L["Bracketed name replacement will take place in guild, officer, party, and raid chat.  All settings are per-profile except for the abbreviations used in rank tags, which are stored for each guild and configurable by all characters."],
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = -1,
				},
			},
		},
		ranktags = {
			name = L["Rank Tags"],
			desc = L["How and when to use them"],
			type = 'group',
			order = 30,
			get = false, set = false,
			args = {
				-- rank-tag-related stuff filled in below
			},
		},
		advanced = {
			name = L["Advanced"],
			desc = L["Unusual options"],
			type = 'group',
			order = 50,
			args = {
				-- Some of these option strings are left unlocalized.
				debug = {
					name = L["Toggle debugging"],
					desc = L["This is a metric assload of text output, especially if HCD is active"],
					type = 'toggle',
					arg  = "HookChat",
					order = 55,
				},
				hcd_debug = {
					name = "Toggle HCD",
					desc = "You really want to put this in its own window. |cffff0000REQUIRES /reload AFTER ENABLING.|r",
					type = 'toggle',
					order = 56,
				},
				print_chatframes = {
					name = "Print Chatframe Numbers",
					desc = "Print each chat window number in its frame, for easy reference in the next slider options",
					type = 'execute',
					--func = print_chatframes filled in below
					order = 57,
				},
				debugging_frame = {
					name = "Debugging Output Chatframe",
					desc = "Which chat window to use for normal debug output",
					type = 'range',
					min  = 1,
					max  = NUM_CHAT_WINDOWS,
					step = 1,
					order = 58,
				},
				hcd_debugging_frame = {
					name = "HCD Debugging Output Chatframe",
					desc = "Which chat window to use for HCD output. |cffff0000REQUIRES /reload AFTER CHANGING.|r",
					type = 'range',
					min  = 1,
					max  = NUM_CHAT_WINDOWS,
					step = 1,
					order = 59,
				},
				max_chatframe_num = {
					name = "Max Chatframe",
					desc = "Highest chatframe number to hook",
					type = 'range',
					min  = 1,
					max  = NUM_CHAT_WINDOWS,
					step = 1,
					arg  = "HookChat",
					order = 60,
				},
				-- If they screw up the input field, validation error messages are lost
				-- http://www.wowace.com/addons/ace3/tickets/2
				alt_pattern_warning = {
					name = [[|cff5500ffATTENTION!|r  If you plan on changing the Pattern field here, you must use the "/oga" command to display this menu instead of Blizzard's "Escape -> Interface -> Addons" menu.  This is due to a bug not under my control.  When it is fixed, this annoying text will be removed.]],
					type = 'description',
					hidden = function() return pattern_editing_safe end,
					cmdHidden = true,
					order = 64,
				},
				alt_pattern = {
					name = L["Pattern"],
					desc = L["Lua pattern to detect and extract |cff00ff00<NAME>|r out of the player and officer note fields (see HOWTO.txt for more)."],
					type = 'input',
					hidden = function() return not pattern_editing_safe end,
					validate = "UpdatePattern",
					order = 65,
				},
			},
		},
		--profiles =   filled in OnInit
	}
}


-----------------------------------------------------------------------------
-- other locals
local assert, next, pairs, ipairs = assert, next, pairs, ipairs
local guildname, guildkey, isofficer = 'NONE', 'NONE', false
local flib = LibStub("LibFarmbuyer", true)
local lua_pattern = "NON-NIL PLACEHOLDER THAT WILL NEVER MATCH BUT WILL ALSO BE INSTANTLY GARBAGE COLLECTED"
local chat_msg_filter_function
local get_colored_name
local event2channame = {
	['CHAT_MSG_GUILD']          = "guild",
	['CHAT_MSG_OFFICER']        = "officer",
	['CHAT_MSG_PARTY']          = "party",
	['CHAT_MSG_PARTY_LEADER']   = "party",
	['CHAT_MSG_RAID']           = "raid",
	['CHAT_MSG_RAID_LEADER']    = "raid",
}

-- The string.format() sequences need to be changed into patterns that would
-- match the corresponding output; the easiest way is by feeding string captures
-- to format() instead of player names.
-- Keys:    modified match pattern
-- Values:  search and replace format patterns
local system_patterns = {
	-- ERR_FRIEND_ONLINE_SS = "|Hplayer:%s|h[%s]|h has come online.";
	--    The first %s becomes the string capture.  To avoid format() choking
	--    on the second %s, we first replace it with something harmless and
	--    simple to match.
	--    The square brackets in the final output must be escaped, so they
	--    get loaded up with '%'s to make it through all the formatting.
	--    |Hplayer:([^:|]+)|h%[.-%]|h has come online.
	[ERR_FRIEND_ONLINE_SS:gsub('%[%%s%]','%%%%[.-%%%%]'):format('([^:|]+)')]
		= { search = "|h%%[%s%%]|h", replace = "|h%%[%s%%]|h" },

	-- ERR_GUILD_LEAVE_S = "%s has left the guild.";
	--    (%S+) has left the guild.
	[ERR_GUILD_LEAVE_S:format('(%S+)')]
		= { search = '%s',  replace = '%s', appendalts = true },

	-- ERR_GUILD_REMOVE_SS = "%s has been kicked out of the guild by %s.";
	--    Similar to "come online".
	--    (%S+) has been kicked out of the guild by .*.
	[ERR_GUILD_REMOVE_SS:format('(%S+)', '.*')]
		= { search = '%s',  replace = '%s', appendalts = true },
}

local state = setmetatable({
	me = UnitName("player"),
	do_rank_tags = false,
	-- on the assumption that all AddMessage's are the same, and if any of them
	-- has been replaced it'd be this one anyhow:
	orig = DEFAULT_CHAT_FRAME.AddMessage,
	author_debug = not not (flib and flib.author_debug),   -- must be boolean
	hcd_debugging_frame = false,
	debugging = function() return addon.db.profile.debug end,
	last_colored_name = false,   -- '|cff........Me|r'
	tag_channels = {},
	frame_data = {},
}, {__index = _G})

-- First argument may optionally be a specific chatframe to use, or a boolean
-- overriding the current debug setting.
state.dprint = function(...)
	local t = "<D>"
	local frame
	local do_it = addon.db.profile.debug
	local first = ...
	if type(first) == 'table' and first.AddMessage then
		frame, first = first, 2
	elseif type(first) == 'boolean' then
		frame, first = addon.db.profile.debugging_frame, 2
		do_it = do_it or first
	else
		frame, first = addon.db.profile.debugging_frame, 1
	end
	if not do_it then return end
	if type(frame) ~= 'table' then
		frame = _G["ChatFrame"..frame]
	end
	for i = first, select('#',...) do
		local x = select(i,...)
		t = t .. " " .. tostring(x)   -- take that, garbage collector...
	end
	t = t:gsub("\124","\124\124")
	state.orig(frame,t,1,0,0)  -- specifically not tailcalling
end

local function print_chatframes()
	for i = 1, NUM_CHAT_WINDOWS do
		local cf = _G["ChatFrame"..i]
		if not cf then break end
		state.orig(cf,"This is frame number " .. i .. ".",1,1,1)
	end
end
addon.options.args.advanced.args.print_chatframes.func = print_chatframes
addon.options.args.general.args.use_alt_rank.desc =
	L["In addition to notes, always include members whose rank includes the word '%s' in some form, and use the entire note field as their main's name."]:format(altrank)


-- Deal with Prat's various possibilities.  If it's not in use, then do our
-- own chat hooks.  If their alt module is fired up, then add our markup pattern
-- (none of its defaults will match).  Otherise hook into their message format
-- function.
--
-- Returns:
--   "normal" if we should keep doing our thing
--   "bail" if we should leave maps in place but not change chat
--   "cleaup" if we should bail and also clear our own memory
--
--local DO = 0
local function handle_prat()
	if not Prat then return "normal" end
	local mod = Prat.Addon:GetModule("AltNames", --[[silent=]]true)
	if not mod then return "normal" end
	if mod:IsEnabled() then
		local addalt = mod.addAlt
		local pattern = "(%S+)%s+" .. lua_pattern
		function mod:addAlt(str,...)
			local alt,main = str:match(pattern)
			if main then return addalt(self, alt..' '..main, ...)
			else return addalt(self,str,...)  -- not one of ours
			end
		end
		return "cleanup"
	end
	local function hook(...)
		--if DO == 2 then return end
		local e,m = ...
		local player = m.PLAYERLINK
		--print(e, "player is", player, "escaped is", player:gsub('\124', '\124\124'))
		if player ~= state.me or true then
			--print('replacing', player)
			local mainname = addon.alts_to_main[player]
			--print('main is', mainname)
			if mainname and mainname ~= "" then
				m.PLAYER = ("%s (%s)"):format(player, mainname)
				--print('m.player now', m.PLAYER)
				--DO = DO + 1
			end
		end
	end
	Prat.RegisterChatEvent(nametag, "Prat_FrameMessage", hook)
	--[==[
	Prat.RegisterChatEvent("OuroGuildAlts", "Prat_PreAddMessage", function(...)
		if DO == 2 then return end
		local e,m,_,_,out = ...
		print(e, out:gsub('\124', '\124\124'))
				if DO == 1 then DO = DO + 1 end
	end)
	]==]
	return "bail"
end


-----------------------------------------------------------------------------
addon = LibStub("AceAddon-3.0"):NewAddon(addon, nametag,
                "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0")

function addon:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("OuroGuildAltsDB", self.defaults, --[[Default=]]true)

	if not next(self.db.global.guilds) then
		self:load_Ouroboros_defaults()
	end
	self.load_Ouroboros_defaults = nil

	for c in self.db.profile.tag_channels:gmatch("[^,]+") do
		state.tag_channels[c] = true
		state.tag_channels[#state.tag_channels+1] = c
	end

	local AceDBOptions = LibStub("AceDBOptions-3.0", true)
	if AceDBOptions then
		self.options.args.profiles = AceDBOptions:GetOptionsTable(self.db)
		self.options.args.profiles.order = 200
	end

	self.options.args.general.args.version.name = 
		"|cff30adffVersion " .. (GetAddOnMetadata(nametag, "Version") or "?") .. "|r"
	LibStub("AceConfig-3.0"):RegisterOptionsTable(nametag, self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(nametag, "Ouro Guild Alts")
	self.optionsFrame.okay = function() pattern_editing_safe = false end
	self.optionsFrame.refresh = self.optionsFrame.okay
	self.optionsFrame.cancel = self.optionsFrame.okay
	self:RegisterChatCommand("oga", "OnChatCommand")

	self:SetEnabledState(self.db.profile.enable)
	self.OnInitialize = nil
end


--[[
function addon:FriendsFrame_Pause()
	FriendsFrame:UnregisterEvent("WHO_LIST_UPDATE")
end
function addon:FriendsFrame_Resume()
	FriendsFrame:RegisterEvent("WHO_LIST_UPDATE")
end
]]


function addon:OnEnable()
	self:RegisterEvent("PLAYER_LOGOUT")
	if not self.db.profile.enable then
		return self:Disable()
	end

	if self.prat == "cleanup" then
		return self:Print(L["ERROR: After turning off Prat modules, you will need to relog or /reload to get this addon working."])
	end

	-- Blizzard's code dispatches a GuildRoster() call 5 seconds after load.
	-- Sometimes these events fire before the data is actually present, leading
	-- to screwups during the scan.  Waiting a little while if possible is better.
	-- Sometimes the events get lost (unknown conditions), so we will force
	-- a scan if it hasn't been done soon enough.
	local maybe = LibStub("AceTimer-3.0",true)  -- only used once, not worth packaging
	if maybe then
		maybe.ScheduleTimer(self, function()
				self:RegisterEvent("GUILD_ROSTER_SHOW", "LoadGuild")
				self:RegisterEvent("GUILD_ROSTER_UPDATE", "LoadGuild")
			end, 2)
		maybe.ScheduleTimer(self, function()
				if not self.hooked then
					self:LoadGuild()
				end
			end, 10)
	else
		-- enh, register now and hope that the lagloop in LoadGuild is enough
		self:RegisterEvent("GUILD_ROSTER_SHOW", "LoadGuild")
		self:RegisterEvent("GUILD_ROSTER_UPDATE", "LoadGuild")
	end
end

function addon:OnDisable()
	self:UnhookChat()
end

function addon:ToggleEnable()
	if self.db.profile.enable then
		self:Enable()
		GuildRoster()
	else
		self:Disable()
	end
end

function addon:PLAYER_LOGOUT()
	--self.db.profile.debug = nil
	self.db.global.guilds.NONE = nil
	for name,gtab in pairs(self.db.global.guilds) do
		if not next(gtab.abbr) then
			table.wipe(self.db.global.guilds[name])
			self.db.global.guilds[name] = nil
		end
	end
	self.db.profile.tag_channels = table.concat(state.tag_channels,',')
end


do
	local optgroup = addon.options.args.ranktags.args

	optgroup.do_rank_tags = {
		name = L["Use rank tags"],
		desc = L["Add short rank-based tags to names"],
		type = 'toggle',
		get = "GetOption", set = "SetOption",
		arg  = "UpdateRankTagging",
		order = 15,
	}

	optgroup.rank_tag_control = {
		name = L["Rank tag details"],
		type = 'group',
		inline = true,
		get = false, set = false,
		disabled = function(i) return not i.handler.db.profile.do_rank_tags end,
		arg  = "UpdateRankTagging",
		order = 20,
		args = {},
	}
	local RTdisabled = optgroup.rank_tag_control.disabled

	optgroup.warn = {
		name = L["Due to Blizzard API limitations, a rank tag for a guildmate will only start taking effect when that guildmate is nearby and (usually) grouped with you.  To conserve memory and CPU cycles, Ouro Guild Alts intentionally makes no special efforts to work around these limitations."],
		type = 'description',
		cmdHidden = true,
		order = 30,
	}

	optgroup = optgroup.rank_tag_control.args

	optgroup.chanlist = {
		name = CHAT_CHANNELS,
		desc = L["Add rank tags in these channels"],
		type = 'multiselect',
		dialogControl = "Dropdown",
		order = 5,
		-- these need to be of function type rather than string keys of members
		values = function(info) return addon:ChannelList() end,
		get = function(info,x) return addon:GetChannel(x) end,
		set = function(info,val) return addon:SetChannel(val) end,
	}

	optgroup.do_rank_tags_officer = {
		name = L["Tags only when officer"],
		desc = L["Use rank tags only when you are a guild officer ('officer' defined as being able to /ginvite)"],
		type = 'toggle',
		get = "GetOption", set = "SetOption",
		arg  = "UpdateRankTagging",
		order = 10,
	}
	optgroup.spacer1 = {
		name = '',
		type = 'description',
		cmdHidden = true,
		width = 'full',
		order = 20,
	}

	local guildlist_v                -- known guild keys
	local workingguild = guildkey    -- current dropdown selection
	local function no_g_selection()
		return workingguild == 'NONE' and true or (not workingguild)
	end

	optgroup.guildlist = {
		name = GUILD,
		desc = L["Which guild to configure"],
		type = 'select',
		order = 21,
		width = 'double',
		values = function(info)
			if not guildlist_v then
				-- Build a list of all previously configured guild-realm combos, plus
				-- the current guild whether it's in the list or not.  (This is the only
				-- way we allow of getting new combos potentially added to the list.)
				guildlist_v = { [guildkey] = guildkey }
				for gk in pairs(addon.db.global.guilds) do
					guildlist_v[gk] = gk
				end
			end
			return guildlist_v
		end,
		get = function(info) return workingguild == 'NONE' and guildkey or workingguild end,
		set = function (info, key) workingguild = key end,
	}

	local workingrank
	local function no_r_selection() return not workingrank end
	optgroup.ranklist = {
		name = GUILDCONTROL_GUILDRANKS, --"guild ranks"
		desc = L["List of ranks for the guild to the left"],
		type = 'select',
		order = 22,
		disabled = function(i) return RTdisabled(i) or no_g_selection() end,
		values = function(info)
			if RTdisabled(info) then return {} end   -- meh
			local g = addon.db.global.guilds
			if not g[workingguild] then
				-- configuring our current guild for the first time
				g[workingguild] = { abbr = {}, ranks = {} }
			end
			g = g[workingguild]
			if (not no_g_selection()) and workingguild == guildkey then
				-- it's our current guild, we can refresh the ranks
				g.ranks = {}
				for i=1, GuildControlGetNumRanks() do
					g.ranks[i-1] = GuildControlGetRankName(i)
				end
			else
				-- must use stored ranks for other guilds
			end
			return g.ranks
		end,
		get = function(info) return workingrank or '' end,
		set = function (info, key) workingrank = key end,
	}

	optgroup.abbredit = {
		name = L["Abbreviation"],
		desc = L["What to append to names for the selected rank"],
		type = 'input',
		order = 23,
		width = 'half',
		disabled = function(i) return RTdisabled(i) or no_r_selection() end,
		get = function(info)
			if no_g_selection() or no_r_selection() then return '' end
			local g,r = addon.db.global.guilds[workingguild], tonumber(workingrank) or 0
			return g.abbr[r]
		end,
		set = function (info, val)
			if no_g_selection() or no_r_selection() then return end
			val = tostring(val)
			if not val then return end
			local g,r = addon.db.global.guilds[workingguild], tonumber(workingrank)
			g.abbr[r] = val ~= '' and val or nil
		end,
	}

	optgroup.spacer2 = {
		name = '',
		type = 'description',
		cmdHidden = true,
		width = 'full',
		order = 40,
	}
	optgroup.clear = {
		name = L["Clear Guild"],
		desc = L["Clear settings for selected guild"],
		type = 'execute',
		order = 42,
		disabled = function(i) return RTdisabled(i) or no_g_selection() end,
		func = function(info)
			-- Yes, they need to be wiped before being nil'd
			-- http://forums.wowace.com/showthread.php?p=307195#post307195
			table.wipe(addon.db.global.guilds[workingguild])
			addon.db.global.guilds[workingguild] = nil
			-- Update the UI as well
			guildlist_v[workingguild] = nil
			workingrank = nil
			info.handler:Printf(L['Configuration for %s%s|r removed.  If this was your current guild, you will need to relog, /reload, or click "%s%s|r" before the change takes effect.'],
				GREEN_FONT_COLOR_CODE, workingguild,
				NORMAL_FONT_COLOR_CODE, L["Rescan Guild Roster"])
			workingguild = 'NONE'
		end,
	}

	function addon:PrepareRankTagging()
		workingguild = guildkey

		self.PrepareRankTagging = nil
	end

	local mt = {
		__index =
			function(t,name)
				state.dprint("rank_lookup starting", name)
				local gn,gr,rank = GetGuildInfo(name)  -- returns 0-based rank
				state.dprint(gn, gr, rank, "and stored guildname is", guildname)
				if gn ~= guildname then return nil end
				rank = addon.rank_letter_map[rank]
				if not rank then return nil end
				t[name] = rank
				state.dprint("rank_lookup returning", name, rank)
				return rank
			end
	}

	function addon:UpdateRankTagging()
		state.do_rank_tags = self.db.profile.do_rank_tags
		if self.db.profile.do_rank_tags_officer then
			state.do_rank_tags = state.do_rank_tags and isofficer
		end
		if state.do_rank_tags and not self.rank_lookup then
			self.rank_lookup = setmetatable({}, mt)
			local rlm = self.db.global.guilds[guildkey]
			rlm = rlm and rlm.abbr
			-- If it doesn't exist, force replacements to silently fail by
			-- using an empty-ish table.  The metatable is empty enough for
			-- this purpose and saves creating a dummy table.
			self.rank_letter_map = rlm or mt
		end
	end
end


do
	local function disp(txt)
		return GREEN_FONT_COLOR_CODE .. txt .. FONT_COLOR_CODE_CLOSE
	end

	function addon:ChannelList()
		-- cribbed from Acheron
		local display_channel_list = {
			["say"] = disp"say",
			["party"] = disp"party",
			["raid"] = disp"raid",
			["guild"] = disp"guild",
			["officer"] = disp"officer",
		}

		local numbered = {}
		for _,chan in ipairs{EnumerateServerChannels()} do
			numbered[chan] = true
		end

		local joined = {GetChannelList()}
		for i=1, #joined, 2 do
			-- don't include the public numbered channels
			if not numbered[ joined[i+1] ] then
				display_channel_list[ tostring(joined[i]) ] = disp(joined[i+1])
			end
		end

		return display_channel_list
	end

	function addon:GetChannel (name)
		return state.tag_channels[name]
	end

	function addon:SetChannel (name)
		local was_on = not not state.tag_channels[name]
		if was_on then
			state.tag_channels[name] = nil
			tDeleteItem (state.tag_channels, name)
		else
			state.tag_channels[name] = true
			state.tag_channels[#state.tag_channels+1] = name
		end
	end
end


-- Thanks to jerry for the nifty arg idea.
function addon:SetOption (info, value)
	local name = info[#info]
	self.db.profile[name] = value
	local arg = info.arg
	if arg then self[arg](self) end
end

function addon:GetOption (info)
	local name = info[#info]
	return self.db.profile[name]
end


do
	-- Try to compensate for AceGUI ticket #2.
	local function err (from_menu, txt)
		if from_menu then
			return txt
		else
			addon:Print(ERROR_CAPS, ": ", txt)
		end
	end
	function addon:UpdatePattern (from_menu, pattern)
		if type(pattern) ~= 'string' then
			return err(from_menu,L["The pattern must be an actual string."])
		end

		local pat, count = pattern:gsub('<NAME>$', '(.-)$')
		if count == 0 then
			pat, count = pattern:gsub('<NAME>', '(.-)')
		end
		if count ~= 1 then
			return err(from_menu,
				(L['The string "<NAME>" must appear exactly once in the pattern "%s", currently %s.  Close and reopen the window to try again.']):format(pattern,tostring(count)))
		end
		lua_pattern = pat
		state.dprint("UpdatePattern arrives at:", lua_pattern)
		return true
	end
end


function addon:OnChatCommand (input)
	pattern_editing_safe = true
	if not input or input:trim() == "" then
		-- running this before the guild roster has been scanned leads to trouble.
		-- here's a crude gate for fast typists:
		if not self.hooked then
			self:LoadGuild()
		end
		LibStub("AceConfigDialog-3.0"):Open(nametag)
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(self, "oga", nametag, input)
	end
end


--[[
Observations:
1) There are a lot of funky ways to screw up text before AddMessage ever sees
   it, some from Blizzard and some by other addons.  This means that debugging
   the pattern matching after each patch can be a long and tedious process,
   made tractable by tons of debugging statements.
2) Testing for the same debug flag over and over on a hot code path is bad,
   especially when it will almost never change.  Debug code needs to either be
   there all the time or none of the time, when possible.
3) When faced with an odd pattern-not-matching situation, we *really* need to
   be able to turn on debugging without asking a user to install a specical
   version with extra print code.

So instead of writing a function, we write the source for that function and
surround the very slow parts with some sentinel tokens.  When starting up for
a game session, the tokens are stripped -- either removing the debugging code
or leaving it behind -- and the result is compiled then.  This is as close to
a macro processor as we can get without using a modified Lua.

Added fun from the fact that loadstring runs in the global env.  WTB Lua 5.2.
]]
function addon:HookChat ()
	if state.debugging() or state.author_debug then
		_G.OGA = self
	end

	self:UnhookChat()

	self.prat = self.prat or handle_prat()
	handle_prat = nil
	if self.prat == "bail" then
		return
	elseif self.prat == "cleanup" then
		self.rank_letter_map = nil
		self.alts_to_main = nil
		self.mains_alts = nil
		self.rank_lookup = nil
		return
	end

	if not self.AddMessage then
		local func_def
		if self.db.profile.hcd_debug then
			state.hcd_debugging_frame = _G["ChatFrame" .. self.db.profile.hcd_debugging_frame]
			func_def = self.func_def:gsub("%-%-<HCD",""):gsub("%-%->HCD","")
		else
			func_def = self.func_def:gsub("%-%-<HCD.-%-%->HCD","")
		end
		local func,err = loadstring(func_def, "Ouro Guild Alts")
		if func then
			setfenv (func, state)
			self.AddMessage = assert(func())   -- don't pcall this for now
			self.func_def = nil
		else
			self:Print(ERROR_CAPS, ": ", err)
			self:Print(L["ERROR: Could not create chat hook functions, giving up."])
			return self:Disable()
		end
	end

	if not get_colored_name then
		-- Delay this to allow all the other addons that hook or outright
		-- replace GCN to do their stuff first.
		get_colored_name = _G.GetColoredName
		_G.GetColoredName = function(...)
			local r = get_colored_name(...)
			state.last_colored_name = r
			-- The only magic characters we're likely to find are the ones
			-- in texture format strings... what else?
			state.last_colored_is_fragile = r:find('[%.%+%-]')
			return r
		end
	end

	--ChatFrame_AddMessageEventFilter ('CHAT_MSG_CHANNEL',        chat_msg_filter_function)
	ChatFrame_AddMessageEventFilter ('CHAT_MSG_GUILD',          chat_msg_filter_function)
	ChatFrame_AddMessageEventFilter ('CHAT_MSG_OFFICER',        chat_msg_filter_function)
	ChatFrame_AddMessageEventFilter ('CHAT_MSG_PARTY',          chat_msg_filter_function)
	ChatFrame_AddMessageEventFilter ('CHAT_MSG_PARTY_LEADER',   chat_msg_filter_function)
	ChatFrame_AddMessageEventFilter ('CHAT_MSG_RAID',           chat_msg_filter_function)
	ChatFrame_AddMessageEventFilter ('CHAT_MSG_RAID_LEADER',    chat_msg_filter_function)
	ChatFrame_AddMessageEventFilter ('CHAT_MSG_SYSTEM',         chat_msg_filter_function)

	if _G.CHAT_TIMESTAMP_FORMAT == nil then
		local n = {}
		for pat, formats in next, system_patterns do
			n['^'..pat] = formats
		end
		system_patterns = n
	end
	--flib.tabledump(system_patterns)

	for i = 1, self.db.profile.max_chatframe_num do
		local cf = _G["ChatFrame"..i]
		if not cf then break end
		self:RawHook (cf, "AddMessage", --[[overridesecure=]]true)
		state.frame_data[i] = {
			legit = false,      -- have we extracted data for this frame+event?
			sender = false,     -- official player name used for lookups
			channel = '',       -- used for deciding rank tagging
			target = '',        -- target string to try and replace
			spattern = '',      -- format() string taking 'target' to search for...
			rpattern = '',      -- ...and replace
		}
	end

	TradeSkillFrame_LoadUI()
	--self:SecureHook ("TradeSkillGuildCraftersFrame_Update")
	if TradeSkillGuildListingMixin then
		--self:SecureHook (TradeSkillGuildListingMixin, "Refresh", "GuildCraftersFrameRefresh")
	else
		self:Print "crap, TradeSkillFrameUI changed again, no ListingMixin table"
	end

	GuildFrame_LoadUI()
	self:SecureHook ("GuildRoster_Update")
	-- can't do this in the XML because GMDF doesn't exist then
	-- eventually move all that into Lua
	OuroGuildAltsRosterDetailAltListFrame:SetParent(GuildMemberDetailFrame)
	OuroGuildAltsRosterDetailAltListFrame:SetPoint("TOPLEFT", GuildMemberDetailFrame, "BOTTOMLEFT", 0, 0)
	OuroGuildAltsRosterDetailAltListFrameTitle:SetText(L["List of alts in guild:"])

	self.hooked = true
end

function addon:UnhookChat()
	self:UnhookAll()
	self.hooked = nil
end


function addon:LoadGuild (a1)
	--self:FriendsFrame_Pause()

	if not IsInGuild() then return end

	-- If it's triggered from a GUILD_ROSTER event, this is the event name and
	-- we might want spam.  If it's a just-in-case call, this is nil and we never
	-- want spam.  If it's a manual click, it's the aceconfig 'info' table with
	-- more args following and we definitely should see spam.
	local whence = ((type(a1) == 'string') and 'LOGIN') or
	               ((type(a1) == 'table') and 'MANUAL') or
	               'INTERNAL'

	local nAlt   = 0
	local nGuild = GetNumGuildMembers(true) or 0

	-- Try and compensate for laggy-ass servers.
	-- http://xkcd.com/37/
	if nGuild == 0 then
		self.lagloop = self.lagloop and (self.lagloop - 1) or 5
		if self.lagloop > 0 then
			return GuildRoster()   -- dispatch a request
		else
			self:Print(L["ERROR: Repeatedly failed to retrieve guild roster, giving up."])
			return self:Disable()
		end
	end
	self.lagloop = nil
	self:UnregisterEvent("GUILD_ROSTER_SHOW")
	self:UnregisterEvent("GUILD_ROSTER_UPDATE")

	-- Build the map
	local GetGuildRosterInfo = _G.GetGuildRosterInfo
	local tinsert, tsort, tconcat = table.insert, table.sort, table.concat

	local actual_toons = {}
	self.alts_to_main = {}
	self.mains_alts = {}
	self:UpdatePattern (nil, self.db.profile.alt_pattern)  -- builds lua_pattern
	local altrank_pattern = '%f[%w]' .. altrank .. '%f[%W]'
	for i = 1, nGuild do
		local cleannote
		local name, rank, rankIndex, level, class, zone, note, officernote, online, status = GetGuildRosterInfo(i)

		actual_toons[name] = true
		if type(officernote) == 'string' and officernote ~= "" then
			cleannote = officernote:match(lua_pattern)
		end
		if (not cleannote) and type(note) == 'string' and note ~= "" then
			cleannote = note:match(lua_pattern)
		end
		if (not cleannote) and self.db.profile.use_alt_rank and rank:lower():match(altrank_pattern) then
			state.dprint("whoa, rank match!", rank, name, note)
			cleannote = note
		end
		if cleannote and cleannote ~= "" and name:lower() ~= cleannote:lower() then
			self.alts_to_main[name] = cleannote
			nAlt = nAlt + 1
			self.mains_alts[cleannote] = self.mains_alts[cleannote] or {}
			tinsert(self.mains_alts[cleannote], name)
		end
	end

	for main,list in pairs(self.mains_alts) do while true do
		if not actual_toons[main] then
			-- somebody had a fake note like "alt(Silly Pirate Druid)"
			self.mains_alts[main] = nil
			break -- 'continue'
		end

		tsort(list)
		self.mains_alts[main] = tconcat (list, ", ")
		break
	end end

	if (whence == 'LOGIN' and self.db.profile.stats_at_login) or
	   (whence == 'MANUAL')
	then
		self:Printf(L["Found %d guild members, %d of whom are labeled as alts."], nGuild, nAlt)
	end
	guildname = GetGuildInfo("player") or 'NONE'
	guildkey = ("<%s> - %s"):format(guildname, GetRealmName())
	isofficer = not not CanGuildInvite()
	self.rank_lookup = nil
	self.rank_letter_map = nil
	if self.PrepareRankTagging then
		self:PrepareRankTagging()
	end
	self:UpdateRankTagging()

	--self:FriendsFrame_Resume()
	if not self.hooked then
		self:HookChat()
	end
end

-- Post-hook from Blizzard_TradeSkillDetails.lua:407 or thereabouts.
function addon:GuildCraftersFrameRefresh()
	local obj = TradeSkillGuildListingMixin
	if obj.waitingOnData then return end

	for i, craftersButton in ipairs(obj.Container.ScrollFrame.buttons) do
		if craftersButton and craftersButton:IsShown() then
			-- The '.fullName' is only set if the player is online.  Unhelpful.
			local name = craftersButton:GetText()
			local main = self.alts_to_main[name]
			if main and self.mains_alts[main] then
				craftersButton:SetText(("%s (%s)"):format(name,main))
			end
		end
	end
end


-- Post-hook from Blizzard_TradeSkillUI.lua:1088 or thereabouts.
function addon:TradeSkillGuildCraftersFrame_Update()
	local button, main
	for i = 1, TRADE_SKILL_GUILD_CRAFTERS_DISPLAYED do
		button = _G['TradeSkillGuildCrafter'..i]
		if button and button:IsShown() and button.name then
			main = self.alts_to_main[button.name]
			if self.mains_alts[main] then
				button:SetText(("%s (%s)"):format(button.name,main))
			end
		end
	end
end


-- Post-hook from Blizzard_GuildRoster.lua:120 or thereabouts.
function addon:GuildRoster_Update()
	local frame = OuroGuildAltsRosterDetailAltListFrame
	local text = OuroGuildAltsRosterDetailAltListFrameList

	frame:Hide()

	local GMDF = assert(GuildMemberDetailFrame)
	if not GMDF:IsShown() then return end

	local current = GuildFrame.selectedName
	if not current then return end

	local list = self.mains_alts[current]
	if not list then return end

	-- Rather than gluing the bottomleft corner to GuildFrame's bottomright, we
	-- force the text to wrap (setwidth on fontstring) and then resize the frame
	-- height to be just a little taller.
	text:SetText(list)
	text:SetWidth(frame:GetRight() - frame:GetLeft()
		- 45)  -- 40 left margin, 5 right margin
	frame:SetHeight(text:GetHeight()
		+ OuroGuildAltsRosterDetailAltListFrameTitle:GetHeight()
		+ 50)
	frame:Show()
end


--[[
	As of 4.2, raw channel text looks like this:
		|Hchannel:CHATGROUP|h[6. foo]|h |Hplayer:Person:lineID:CHATGROUP|h[Person]|h: hello
	and displays like
		[6. foo] [Person]: hello
	The CHATGROUP is "GUILD", "OFFICER", "PARTY", "RAID", or "CHANNEL:n" with
	n == channel number.  Party leaders/guides and raid leaders are folded into
	that same scheme.

	For system messages, the ":lineID" is not present:
		|Hplayer:Person|h[Person]|h has come online.

	Colorized raid line looks like
		|Hchannel:RAID|h[Raid]|h |Hplayer:Me:lineID:RAID|h[|cffxxxxxxMe|r]|h: hello

	If timestamps are in use, they're just tacked onto the beginning of the whole thing.
]]


-- Does not actually "filter" out anything, ever.  But we can collect data
-- about the AddMessage call that's about to happen.
function chat_msg_filter_function (frame, event, msg, ...)
	-- Normally never happens, but it turns out oRA3's readycheck "frame"
	-- gets fed here as well, so might as well handle all the weirdness.
	if (frame == nil) or (event == nil) then return false end

	local fd = state.frame_data[frame:GetID()]
	if not fd then return false end
	local arg2sender = ...

	--local t = type(arg2sender)
	--state.orig (frame, GetTime(), "from", arg2sender.."/"..t, "in frame", frame:GetID(), ":", msg, 0.8,1,0.8)
	--state.dprint(state.hcd_debugging_frame, frame:GetName(), "<" .. (fd.channel or "") .. ">:", msg, ...)

	if event == 'CHAT_MSG_CHANNEL' then


	elseif event == 'CHAT_MSG_SYSTEM' then
		-- Blizz does no transforms here; arg1 goes straight out.
		local name, pat, formats
		for p,f in next, system_patterns do
			name = msg:match(p)
			if name then
				pat, formats = p, f
				break
			end
		end
		if name then
			fd.legit = true
			fd.sender = name
			fd.channel = 'system'
			fd.target = name
			fd.spattern = formats.search
			fd.rpattern = formats.replace
			fd.append_text = formats.appendalts and addon.mains_alts[name] and
				("  "..L["List of alts remaining in guild:"].."  "..addon.mains_alts[name])
		end

	elseif arg2sender and arg2sender ~= "" then
		fd.legit = true
		fd.sender = arg2sender
		fd.channel = assert(event2channame[event])
		fd.target = false   -- use the results of GetColoredName, about to be called
		-- I really want to tweak these next two based on the return value
		-- from GetColoredName, which hasn't been called yet.  Sigh.
		fd.spattern = "|h%%[%s%%]|h"
		fd.rpattern = "|h%%[%s%%]|h"

	end

	state.last_colored_is_fragile = nil
	return false
end


addon.func_def = [=[
return function (self, frame, msg, ...)
	if not self.db.profile.enable then
		-- We shouldn't even be here today!
		self:Disable()
		return self.hooks[frame].AddMessage(frame,msg,...)
	end

	local fd = frame_data[frame:GetID()]
	if not fd then
		dprint("hooked AM fired on frame", frame.GetID and frame:GetID() or 'nil', ", past max_chatframe_num!  disabling")
		self:Disable()
		return self.hooks[frame].AddMessage(frame,msg,...)
	end

	if not fd.legit then
		return orig(frame,msg,...)
	end
	fd.legit = false

	local channel = fd.channel
	local sender, replace_target = fd.sender, fd.target or last_colored_name
	--<HCD
	dprint(hcd_debugging_frame, frame:GetName(), "<"..(channel or "").."> sender and replace:", sender, replace_target)
	dprint(hcd_debugging_frame, frame:GetName(), "<"..(channel or "")..">:", msg, ...)
	-->HCD

	if (sender == me
		--<HCD
		and not debugging()
		-->HCD
		) or (not sender)
	then
		return orig(frame,msg,...)
	end

	-- Try to look it up
	local mainname = self.alts_to_main[sender]
	local replacement

	if mainname then
		replacement = replace_target .. " (" .. mainname .. ")"
	end

	if do_rank_tags and (tag_channels[channel]
						 or debugging()
						)
	then
		local r = self.rank_lookup[sender]
		dprint("rank lookup on", sender, "found", r or 'nil')
		if r then
			replacement = (replacement or replace_target) .. ';' .. r
		end
	end

	if replacement then
		local searchfor, replacewith
		if last_colored_is_fragile then  -- this is such a kludge
			--<HCD
			dprint("FRAGILE!", last_colored_is_fragile)
			-->HCD
			searchfor = replace_target:gsub('(%W)','%%%1')
			replacewith = replacement
		else
			searchfor = fd.spattern:format(replace_target)
			replacewith = fd.rpattern:format(replacement)
		end
		--<HCD
		dprint(hcd_debugging_frame, "target:", replace_target, "/", searchfor)
		dprint(hcd_debugging_frame, "replacement:", replacement, "/", replacewith)
		-->HCD
		msg = msg:gsub (searchfor, replacewith)
	end

	if fd.append_text then
		msg = msg .. fd.append_text
		fd.append_text = nil
	end

	return orig(frame,msg,...)
end
]=]--):gsub('@([%u])@',at_strings)

-- vim:noet
